from omnitools import args, utf8d
from .utils import recv_all
from typing import *
import socket
import pickle
import json


class SC:
    def __init__(self, *, host: str = "127.199.71.10", port: int = 39291) -> None:
        self.__s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__s.connect((host, int(port)))

    def request(self, *, command: str, data: Tuple[Tuple, Dict] = args()):
        request = dict(command=command, data=data)
        try:
            request = json.dumps(request).encode()
        except:
            request = pickle.dumps(request)
        import struct
        self.__s.send(struct.pack('>I', len(request)) + request)
        response = recv_all(self.__s)
        try:
            return json.loads(utf8d(response))
        except UnicodeDecodeError:
            return pickle.loads(response)

    def close(self):
        self.__s.close()




from unencryptedsocket import *


def test(data):
    return "Data:\t{}, len: {}".format(data, len(data))


functions = dict(test_command=test)
SS(functions=functions).start()
print("test socket server started.", flush=True)




import struct
import socket


def recv_all(conn: socket.socket) -> bytes:
    length = conn.recv(4)
    if length == b"":
        return b""
    length = struct.unpack('>I', length)[0]
    content = b""
    while len(content) < length:
        content += conn.recv(length-len(content))
        if not content:
            break
    return content





from omnitools import encryptedsocket_function, p, utf8d, dt2yyyymmddhhmmss, debug_info
from .utils import recv_all
import threading
import socket
import struct
import pickle
import json


class SS:
    def __init__(
            self, *, functions: encryptedsocket_function = None,
            host: str = "127.199.71.10", port: int = 39291,
            silent: bool = False, backlog: int = 5
    ):
        self.terminate = False
        self.silent = silent
        self.__s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__s.bind((host, int(port)))
        self.__s.listen(backlog)
        self.functions = functions or {}

    def handler(self, conn: socket.socket, addr: tuple) -> None:
        uid = "{}:{}".format(*addr)
        if not self.silent:
            p("{}\tconnected\t{}".format(dt2yyyymmddhhmmss(), uid))
        try:
            while True:
                request = recv_all(conn)
                if not request:
                    break
                try:
                    request = json.loads(utf8d(request))
                except:
                    request = pickle.loads(request)
                try:
                    if request["command"] in self.functions:
                        response = self.functions[request["command"]](*request["data"][0], **request["data"][1])
                    else:
                        raise Exception("request command '{}' is not in socket functions".format(request["command"]))
                except:
                    response = debug_info()[0]
                try:
                    response = json.dumps(response).encode()
                except TypeError:
                    response = pickle.dumps(response)
                conn.sendall(struct.pack('>I', len(response))+response)
        except:
            p(debug_info()[0])
        finally:
            conn.close()
            if not self.silent:
                p("{}\tdisconnected\t{}".format(dt2yyyymmddhhmmss(), uid))

    def start(self) -> None:
        try:
            while not self.terminate:
                conn, addr = self.__s.accept()
                threading.Thread(target=self.handler, args=(conn, addr)).start()
        except Exception as e:
            if not self.terminate:
                raise e

    def stop(self) -> bool:
        self.terminate = True
        self.__s.close()
        return True




from unencryptedsocket import *
import threading


ps = []
for i in range(10):
    def job(i):
        print(SC().request(command="test_command", data=args("Hello, {}!".format(i))))
    p = threading.Thread(target=job, args=(i,))
    ps.append(p)
    p.start()
for p in ps:
    p.join()

# will fail running in parallel
sc = SC()
# best to use template like:
# sc = lambda: SC(host=..., port=...)
# response = sc().request(...)
long_data = "Hello, 999!"*999
response = sc.request(command="test_command", data=args(long_data))
print(response, len(long_data))
sc.s.close()
print("test socket client started.", flush=True)
